searchData={"items":[{"type":"module","title":"TypedStructor","doc":"","ref":"TypedStructor.html"},{"type":"module","title":"Installation - TypedStructor","doc":"Add `:typed_structor` to the list of dependencies in `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:typed_structor, \"~> 0.1.2\"}\n  ]\nend\n```\n\nAdd `:typed_structor` to your `.formatter.exs` file\n\n```elixir\n[\n  # import the formatter rules from `:typed_structor`\n  import_deps: [..., :typed_structor],\n  inputs: [...]\n]\n```","ref":"TypedStructor.html#module-installation"},{"type":"module","title":"Usage - TypedStructor","doc":"","ref":"TypedStructor.html#module-usage"},{"type":"module","title":"General usage - TypedStructor","doc":"To define a struct with types, use `TypedStructor`,\nand then define fields under the `TypedStructor.typed_structor/2` macro,\nusing the `TypedStructor.field/3` macro to define each field.\n\n```elixir\ndefmodule User do\n  # use TypedStructor to import the `typed_structor` macro\n  use TypedStructor\n\n  typed_structor do\n    # Define each field with the `field` macro.\n    field :id, pos_integer()\n\n    # set a default value\n    field :name, String.t(), default: \"Unknown\"\n\n    # enforce a field\n    field :age, non_neg_integer(), enforce: true\n  end\nend\n```\nThis is equivalent to:\n```elixir\ndefmodule User do\n  defstruct [:id, :name, :age]\n\n  @type t() :: %__MODULE__{\n    id: pos_integer() | nil,\n    # Note: The 'name' can not be nil, for it has a default value.\n    name: String.t(),\n    age: non_neg_integer()\n  }\nend\n```\nCheck `TypedStructor.typed_structor/2` and `TypedStructor.field/3` for more information.","ref":"TypedStructor.html#module-general-usage"},{"type":"module","title":"Options - TypedStructor","doc":"You can also generate an `opaque` type for the struct,\neven changing the type name:\n\n```elixir\ndefmodule User do\n  use TypedStructor\n\n  typed_structor type_kind: :opaque, type_name: :profile do\n    field :id, pos_integer()\n    field :name, String.t()\n    field :age, non_neg_integer()\n  end\nend\n```\nThis is equivalent to:\n```elixir\ndefmodule User do\n  use TypedStructor\n\n  defstruct [:id, :name, :age]\n\n  @opaque profile() :: %__MODULE__{\n    id: pos_integer() | nil,\n    name: String.t() | nil,\n    age: non_neg_integer() | nil\n  }\nend\n```\n\nType parameters also can be defined:\n```elixir\ndefmodule User do\n  use TypedStructor\n\n  typed_structor do\n    parameter :id\n    parameter :name\n\n    field :id, id\n    field :name, name\n    field :age, non_neg_integer()\n  end\nend\n```\nbecomes:\n```elixir\ndefmodule User do\n  @type t(id, name) :: %__MODULE__{\n    id: id | nil,\n    name: name | nil,\n    age: non_neg_integer() | nil\n  }\n\n  defstruct [:id, :name, :age]\nend\n```\n\nIf you prefer to define a struct in a submodule, pass the `module` option.\n```elixir\ndefmodule User do\n  use TypedStructor\n\n  # `%User.Profile{}` is generated\n  typed_structor module: Profile do\n    field :id, pos_integer()\n    field :name, String.t()\n    field :age, non_neg_integer()\n  end\nend\n```\n\nYou can define the type only without defining the struct,\nit is useful when the struct is defined by another library(like `Ecto.Schema`).\n```elixir\ndefmodule User do\n  use Ecto.Schema\n  use TypedStructor\n\n  typed_structor define_struct: false do\n    field :id, pos_integer()\n    field :name, String.t()\n    field :age, non_neg_integer(), default: 0\n  end\n\n  schema \"users\" do\n    field :name, :string\n    field :age, :integer, default: 0\n  end\nend\n```","ref":"TypedStructor.html#module-options"},{"type":"module","title":"Documentation - TypedStructor","doc":"To add a `@typedoc` to the struct type, just add the attribute in the typed_structor block:\n\n```elixir\ntyped_structor do\n  @typedoc \"A typed user\"\n\n  field :id, pos_integer()\n  field :name, String.t()\n  field :age, non_neg_integer()\nend\n```\nYou can also document submodules this way:\n\n```elixir\ntypedstruct module: Profile do\n  @moduledoc \"A user profile struct\"\n  @typedoc \"A typed user profile\"\n\n  field :id, pos_integer()\n  field :name, String.t()\n  field :age, non_neg_integer()\nend\n```","ref":"TypedStructor.html#module-documentation"},{"type":"module","title":"Plugins - TypedStructor","doc":"`TypedStructor` offers a plugin system to enhance functionality.\nFor details on creating a plugin, refer to the `TypedStructor.Plugin` module.\n\nHere is a example of `TypedStructor.Plugins.Accessible` plugin to define `Access` behavior for the struct.\n```elixir\ndefmodule User do\n  use TypedStructor\n\n  typed_structor do\n    plugin TypedStructor.Plugins.Accessible\n\n    field :id, pos_integer()\n    field :name, String.t()\n    field :age, non_neg_integer()\n  end\nend\n\nuser = %User{id: 1, name: \"Phil\", age: 20}\nget_in(user, [:name]) # => \"Phil\"\n```","ref":"TypedStructor.html#module-plugins"},{"type":"macro","title":"TypedStructor.field/3","doc":"Defines a field in a `typed_structor`.","ref":"TypedStructor.html#field/3"},{"type":"macro","title":"Example - TypedStructor.field/3","doc":"# A field named :example of type String.t()\n    field :example, String.t()","ref":"TypedStructor.html#field/3-example"},{"type":"macro","title":"Options - TypedStructor.field/3","doc":"* `:default` - sets the default value for the field\n  * `:enforce` - if set to true, enforces the field and makes its type\n    non-nullable","ref":"TypedStructor.html#field/3-options"},{"type":"macro","title":"TypedStructor.parameter/1","doc":"Defines a type parameter in a `typed_structor`.","ref":"TypedStructor.html#parameter/1"},{"type":"macro","title":"Example - TypedStructor.parameter/1","doc":"# A type parameter named int\n    parameter :int\n\n    fied :number, int # not int()","ref":"TypedStructor.html#parameter/1-example"},{"type":"macro","title":"TypedStructor.plugin/2","doc":"Registers a plugin for the currently defined struct.","ref":"TypedStructor.html#plugin/2"},{"type":"macro","title":"Example - TypedStructor.plugin/2","doc":"typed_structor do\n      plugin MyPlugin\n\n      field :string, String.t()\n    end\n\nFor more information on how to define your own plugins, please see\n`TypedStructor.Plugin`. To use a third-party plugin, please refer directly to\nits documentation.","ref":"TypedStructor.html#plugin/2-example"},{"type":"macro","title":"TypedStructor.typed_structor/2","doc":"Defines a struct with type information.\n\nInside a `typed_structor` block, you can define fields with the `field/3` macro.","ref":"TypedStructor.html#typed_structor/2"},{"type":"macro","title":"Options - TypedStructor.typed_structor/2","doc":"* `:module` - if provided, the struct will be defined in the given module.\n  * `:enforce` - if `true`, the struct will enforce the keys.\n  * `:define_struct` - if `false`, the struct will not be defined. Defaults to `true`.\n  * `:type_kind` - the kind of type to use for the struct. Defaults to `type`, can be `opaque` or `typep`.\n  * `:type_name` - the name of the type to use for the struct. Defaults to `t`.","ref":"TypedStructor.html#typed_structor/2-options"},{"type":"macro","title":"Examples - TypedStructor.typed_structor/2","doc":"defmodule MyStruct do\n      use TypedStructor\n\n      typed_structor do\n        field :name, String.t()\n        field :age, integer()\n      end\n    end\n\nCreates the struct in a submodule instead:\n\n    defmodule MyStruct do\n      use TypedStructor\n\n      typed_structor module: Struct do\n        field :name, String.t()\n        field :age, integer()\n      end\n    end\n\nTo add a `@typedoc` to the struct type and `@moduledoc` to the submodule,\njust add the module attribute in the `typed_structor` block:\n\n    defmodule MyStruct do\n      use TypedStructor\n\n      typed_structor module: Struct do\n        @typedoc \"A typed struct\"\n        @moduledoc \"A submodule\"\n\n        field :name, String.t()\n        field :age, integer()\n      end\n    end","ref":"TypedStructor.html#typed_structor/2-examples"},{"type":"module","title":"TypedStructor.Definition","doc":"The definition struct that holds the `TypedStructor` options,\nfields and parameters.","ref":"TypedStructor.Definition.html"},{"type":"type","title":"TypedStructor.Definition.t/0","doc":"","ref":"TypedStructor.Definition.html#t:t/0"},{"type":"behaviour","title":"TypedStructor.Plugin","doc":"This module defines the plugin behaviour for `TypedStructor`.","ref":"TypedStructor.Plugin.html"},{"type":"behaviour","title":"Plugin Behaviour - TypedStructor.Plugin","doc":"A plugin is a module that implements the `TypedStructor.Plugin` behaviour.\nThree macro callbacks are available for injecting code at different stages:\n\n* `c:init/1`: This macro callback is called when the plugin is used.\n* `c:before_definition/2`: This macro callback is called right before defining the struct.\nNote hat plugins will run in the order they are registered.\n* `c:after_definition/2`: This macro callback is called right after defining the struct.\nNote that plugins will run in the **reverse** order they are registered.","ref":"TypedStructor.Plugin.html#module-plugin-behaviour"},{"type":"behaviour","title":"Example - TypedStructor.Plugin","doc":"Let's define a plugin that defines `Ecto.Schema` while defining a typed struct.\nThis plugin takes a `:source` option which passing to `Ecto.Schema.schema/2`,\nyou can use `belongs_to` and `has_many` directly in the module.\nIt would be used like this:\n```elixir\ndefmodule MyApp do\n  use TypedStructor\n\n  # fix aliases\n  alias __MODULE__.User\n  alias __MODULE__.Post\n\n  typed_structor module: User do\n    # import the plugin with source option\n    plugin EctoSchemaPlugin, source: \"users\"\n\n    field :name, :string, enforce: true\n    field :age, :integer, default: 0\n    # pass redact option to Ecto.Schema.field/3\n    field :password, :string, redact: true\n    has_many :posts, Post\n  end\n\n  typed_structor module: Post do\n    # import the plugin with source option\n    plugin EctoSchemaPlugin, source: \"posts\"\n\n    field :title, :string, enforce: true\n    field :content, :string, enforce: true\n    belongs_to :user, User, enforce: true\n  end\nend\n```\nAfter compiled, you got:\n```elixir\niex> t MyApp.User\n@type t() :: %MyApp.User{\n        __meta__: Ecto.Schema.Metadata.t(),\n        age: integer() | nil,\n        id: integer(),                                                                                                                                                                       name: String.t(),\n        password: String.t() | nil,\n        posts: [MyApp.Post.t()] | nil\n      }\n\niex> t MyApp.Post\n@type t() :: %MyApp.Post{\n        __meta__: Ecto.Schema.Metadata.t(),\n        content: String.t(),\n        id: integer(),\n        title: String.t(),\n        user: MyApp.User.t(),\n        user_id: integer()\n      }\n\niex> MyApp.User.__schema__(:redact_fields)\n[:password]\n\niex> MyApp.User.__schema__(:association, :posts)\n%Ecto.Association.Has{\n  cardinality: :many,\n  field: :posts,\n  owner: MyApp.User,\n  related: MyApp.Post,\n  owner_key: :id,\n  related_key: :user_id,\n  on_cast: nil,\n  queryable: MyApp.Post,\n  on_delete: :nothing,                                                                                                                                                                 on_replace: :raise,\n  where: [],\n  unique: true,\n  defaults: [],\n  relationship: :child,\n  ordered: false,\n  preload_order: []\n}\n\niex> MyApp.Post.__schema__(:association, :user)\n%Ecto.Association.BelongsTo{\n  field: :user,\n  owner: MyApp.Post,\n  related: MyApp.User,\n  owner_key: :user_id,\n  related_key: :id,\n  queryable: MyApp.User,\n  on_cast: nil,\n  on_replace: :raise,\n  where: [],\n  defaults: [],\n  cardinality: :one,\n  relationship: :parent,\n  unique: true,\n  ordered: false\n}\n```\nFollowing is the implementation of the plugin:\n```elixir\ndefmodule EctoSchemaPlugin do\n  use TypedStructor.Plugin\n\n  @impl TypedStructor.Plugin\n  defmacro init(opts) do\n    quote do\n      unless Keyword.has_key?(unquote(opts), :source) do\n        raise \"The `:source` option is not provided.\"\n      end\n\n      # import association functions to the module,\n      # so that we can use `has_many` and `belongs_to` directly\n      import unquote(__MODULE__), only: [has_many: 2, belongs_to: 3]\n    end\n  end\n\n  @impl TypedStructor.Plugin\n  defmacro before_definition(definition, _opts) do\n    # manipulate the definition before defining the struct\n    quote do\n      unquote(definition)\n      # disable defining struct, for Ecto.Schema will define it\n      |> Map.update!(:options, &Keyword.put(&1, :define_struct, false))\n      |> Map.update!(:fields, fn fields ->\n        Enum.flat_map(fields, fn field ->\n          {ecto_type, options} = Keyword.pop!(field, :type)\n          type = unquote(__MODULE__).__ecto_type_to_type__(ecto_type)\n\n          field = Keyword.merge(options, type: type, ecto_type: ecto_type)\n\n          case ecto_type do\n            {:belongs_to, name} ->\n              foreign_key_name =\n                name\n                |> Macro.expand(__ENV__)\n                |> Module.split()\n                |> List.last()\n                |> Macro.underscore()\n                |> Kernel.<>(\"_id\")\n                |> String.to_atom()\n\n              foreign_key =\n                Keyword.merge(options, name: foreign_key_name, type: quote(do: integer()))\n\n              [foreign_key, field]\n\n            _other ->\n              [field]\n          end\n        end)\n      end)\n      |> Map.update!(\n        :fields,\n        &[\n          [name: :__meta__, type: quote(do: Ecto.Schema.Metadata.t()), enforce: true],\n          [name: :id, type: quote(do: integer()), enforce: true]\n          | &1\n        ]\n      )\n    end\n  end\n\n  @impl TypedStructor.Plugin\n  defmacro after_definition(definition, opts) do\n    # here we define the Ecto.Schema\n    quote bind_quoted: [definition: definition, opts: opts] do\n      use Ecto.Schema\n\n      source = Keyword.fetch!(opts, :source)\n\n      schema source do\n        for options <- definition.fields do\n          {name, options} = Keyword.pop!(options, :name)\n          {ecto_type, options} = Keyword.pop(options, :ecto_type)\n          options = Keyword.take(options, [:primary_key, :default, :redact])\n\n          case ecto_type do\n            nil ->\n              # skip some fields\n              nil\n\n            {:has_many, module} ->\n              module = Macro.expand(module, __ENV__)\n\n              Ecto.Schema.has_many(name, module, options)\n\n            {:belongs_to, module} ->\n              module = Macro.expand(module, __ENV__)\n\n              Ecto.Schema.belongs_to(name, module, options)\n\n            _ ->\n              Ecto.Schema.field(name, ecto_type, options)\n          end\n        end\n      end\n    end\n  end\n\n  defmacro has_many(name, queryable) do\n    quote do\n      field unquote(name), {:has_many, unquote(queryable)}\n    end\n  end\n\n  defmacro belongs_to(name, queryable, opts) do\n    quote do\n      field unquote(name), {:belongs_to, unquote(queryable)}, unquote(opts)\n    end\n  end\n\n  def __ecto_type_to_type__(:string), do: quote(do: String.t())\n  def __ecto_type_to_type__(:integer), do: quote(do: integer())\n  def __ecto_type_to_type__({:has_many, module}), do: quote(do: [unquote(module).t()])\n  def __ecto_type_to_type__({:belongs_to, module}), do: quote(do: unquote(module).t())\nend\n```","ref":"TypedStructor.Plugin.html#module-example"},{"type":"macrocallback","title":"TypedStructor.Plugin.after_definition/2","doc":"This macro callback is called right after defining the struct.\n\nIt receives the definition of the struct and the plugin options,\nand its return value is ignored.","ref":"TypedStructor.Plugin.html#c:after_definition/2"},{"type":"macrocallback","title":"TypedStructor.Plugin.before_definition/2","doc":"This macro callback is called right before defining the struct.\n\nIt receives the definition of the struct and the plugin options,\nand it should return the `TypedStructor.Definition` struct or\na list which contains exactly one `TypedStructor.Definition` struct.","ref":"TypedStructor.Plugin.html#c:before_definition/2"},{"type":"macrocallback","title":"TypedStructor.Plugin.init/1","doc":"This macro callback is called when the plugin is used.\n\nHere you can define module attributes, import modules, etc.","ref":"TypedStructor.Plugin.html#c:init/1"},{"type":"module","title":"TypedStructor.Plugins.Accessible","doc":"This plugin implements the `Access` behavior for the struct\nby delegating the `fetch/2`, `get_and_update/3`, and `pop/2`\nfunctions to the `Map` module.\n\n> #### Destructive operations  {: .warning}\n> These operations are not allowed for the struct:\n> * update `:__struct__` key\n> * pop a key\n>\n> The functions will raise an `ArgumentError` if called.\n> To enable these functionalities, override the `get_and_update/3` and `pop/2` functions.","ref":"TypedStructor.Plugins.Accessible.html"},{"type":"module","title":"Usage - TypedStructor.Plugins.Accessible","doc":"```elixir\n  typed_structor do\n    plugin TypedStructor.Plugins.Accessible\n\n    # fields\n  end\n```","ref":"TypedStructor.Plugins.Accessible.html#module-usage"},{"type":"extras","title":"TypedStructor","doc":"# TypedStructor\n\n[![Build Status](https://github.com/elixir-typed-structor/typed_structor/actions/workflows/elixir.yml/badge.svg)](https://github.com/elixir-typed-structor/typed_structor/actions/workflows/elixir.yml)\n[![Hex.pm](https://img.shields.io/hexpm/v/typed_structor.svg)](https://hex.pm/packages/typed_structor)\n[![Documentation](https://img.shields.io/badge/documentation-gray)](https://hexdocs.pm/typed_structor/)\n\n`TypedStructor` is a library for defining structs with types effortlessly.\n(This library is a rewritten version of [TypedStruct](https://github.com/ejpcmac/typed_struct) because it is no longer actively maintained.)\n\n<!-- MODULEDOC -->","ref":"readme.html"},{"type":"extras","title":"Installation - TypedStructor","doc":"Add `:typed_structor` to the list of dependencies in `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:typed_structor, \"~> 0.1.2\"}\n  ]\nend\n```\n\nAdd `:typed_structor` to your `.formatter.exs` file\n\n```elixir\n[\n  # import the formatter rules from `:typed_structor`\n  import_deps: [..., :typed_structor],\n  inputs: [...]\n]\n```","ref":"readme.html#installation"},{"type":"extras","title":"Usage - TypedStructor","doc":"","ref":"readme.html#usage"},{"type":"extras","title":"General usage - TypedStructor","doc":"To define a struct with types, use `TypedStructor`,\nand then define fields under the `TypedStructor.typed_structor/2` macro,\nusing the `TypedStructor.field/3` macro to define each field.\n\n```elixir\ndefmodule User do\n  # use TypedStructor to import the `typed_structor` macro\n  use TypedStructor\n\n  typed_structor do\n    # Define each field with the `field` macro.\n    field :id, pos_integer()\n\n    # set a default value\n    field :name, String.t(), default: \"Unknown\"\n\n    # enforce a field\n    field :age, non_neg_integer(), enforce: true\n  end\nend\n```\nThis is equivalent to:\n```elixir\ndefmodule User do\n  defstruct [:id, :name, :age]\n\n  @type t() :: %__MODULE__{\n    id: pos_integer() | nil,\n    # Note: The 'name' can not be nil, for it has a default value.\n    name: String.t(),\n    age: non_neg_integer()\n  }\nend\n```\nCheck `TypedStructor.typed_structor/2` and `TypedStructor.field/3` for more information.","ref":"readme.html#general-usage"},{"type":"extras","title":"Options - TypedStructor","doc":"You can also generate an `opaque` type for the struct,\neven changing the type name:\n\n```elixir\ndefmodule User do\n  use TypedStructor\n\n  typed_structor type_kind: :opaque, type_name: :profile do\n    field :id, pos_integer()\n    field :name, String.t()\n    field :age, non_neg_integer()\n  end\nend\n```\nThis is equivalent to:\n```elixir\ndefmodule User do\n  use TypedStructor\n\n  defstruct [:id, :name, :age]\n\n  @opaque profile() :: %__MODULE__{\n    id: pos_integer() | nil,\n    name: String.t() | nil,\n    age: non_neg_integer() | nil\n  }\nend\n```\n\nType parameters also can be defined:\n```elixir\ndefmodule User do\n  use TypedStructor\n\n  typed_structor do\n    parameter :id\n    parameter :name\n\n    field :id, id\n    field :name, name\n    field :age, non_neg_integer()\n  end\nend\n```\nbecomes:\n```elixir\ndefmodule User do\n  @type t(id, name) :: %__MODULE__{\n    id: id | nil,\n    name: name | nil,\n    age: non_neg_integer() | nil\n  }\n\n  defstruct [:id, :name, :age]\nend\n```\n\nIf you prefer to define a struct in a submodule, pass the `module` option.\n```elixir\ndefmodule User do\n  use TypedStructor\n\n  # `%User.Profile{}` is generated\n  typed_structor module: Profile do\n    field :id, pos_integer()\n    field :name, String.t()\n    field :age, non_neg_integer()\n  end\nend\n```\n\nYou can define the type only without defining the struct,\nit is useful when the struct is defined by another library(like `Ecto.Schema`).\n```elixir\ndefmodule User do\n  use Ecto.Schema\n  use TypedStructor\n\n  typed_structor define_struct: false do\n    field :id, pos_integer()\n    field :name, String.t()\n    field :age, non_neg_integer(), default: 0\n  end\n\n  schema \"users\" do\n    field :name, :string\n    field :age, :integer, default: 0\n  end\nend\n```","ref":"readme.html#options"},{"type":"extras","title":"Documentation - TypedStructor","doc":"To add a `@typedoc` to the struct type, just add the attribute in the typed_structor block:\n\n```elixir\ntyped_structor do\n  @typedoc \"A typed user\"\n\n  field :id, pos_integer()\n  field :name, String.t()\n  field :age, non_neg_integer()\nend\n```\nYou can also document submodules this way:\n\n```elixir\ntypedstruct module: Profile do\n  @moduledoc \"A user profile struct\"\n  @typedoc \"A typed user profile\"\n\n  field :id, pos_integer()\n  field :name, String.t()\n  field :age, non_neg_integer()\nend\n```","ref":"readme.html#documentation"},{"type":"extras","title":"Plugins - TypedStructor","doc":"`TypedStructor` offers a plugin system to enhance functionality.\nFor details on creating a plugin, refer to the `TypedStructor.Plugin` module.\n\nHere is a example of `TypedStructor.Plugins.Accessible` plugin to define `Access` behavior for the struct.\n```elixir\ndefmodule User do\n  use TypedStructor\n\n  typed_structor do\n    plugin TypedStructor.Plugins.Accessible\n\n    field :id, pos_integer()\n    field :name, String.t()\n    field :age, non_neg_integer()\n  end\nend\n\nuser = %User{id: 1, name: \"Phil\", age: 20}\nget_in(user, [:name]) # => \"Phil\"\n```","ref":"readme.html#plugins"}],"content_type":"text/markdown","producer":{"name":"ex_doc","version":[48,46,51,52,46,49]}}